unit CSA;
interface

uses
  Windows,
  DeviceLevel;

function ADC_InitFastITScanV(DeviceIndex: LongWord): LongWord; cdecl;
function ADC_ResetFastITScanV(DeviceIndex: LongWord): LongWord; cdecl;
function ADC_SetFastITScanVChannels(DeviceIndex, NewChannels: LongWord): LongWord; cdecl;
function ADC_GetFastITScanV(DeviceIndex: LongWord; pBuf: PDouble): LongWord; cdecl;
function ADC_GetITScanV(DeviceIndex: LongWord; pBuf: PDouble): LongWord; cdecl;
function DIO_CSA_DoSync(DeviceIndex: LongWord; var BaseRateHz, DurAms, DurBms, DurCms: Double): LongWord; cdecl;
function DIO_CSA_DebounceSet(DeviceIndex, DebounceCounts: LongWord): LongWord; cdecl;
function DIO_CSA_DebounceReadAll(DeviceIndex: LongWord; pData: Pointer): LongWord; cdecl;
function DAC_CSA_SetRangeLimits(DeviceIndex: LongWord; pData: Pointer): LongWord; cdecl;
function DAC_CSA_ClearRangeLimits(DeviceIndex: LongWord): LongWord; cdecl;
function DIO_CSA_ReadAllString(DeviceIndex: LongWord; pData: Pointer; pDataBytes: PLongWord): LongWord; cdecl;

function ADC_CSA_InitFastScanV(DeviceIndex: LongWord): LongWord; cdecl;
function ADC_CSA_ResetFastScanV(DeviceIndex: LongWord): LongWord; cdecl;
function ADC_CSA_GetFastScanV(DeviceIndex: LongWord; pBuf: PDouble): LongWord; cdecl;

function DACDIO_WriteAll(DeviceIndex: LongWord; pDACCounts: PWord; pDIOData: PByte): LongWord; cdecl;

function DIO_SPI_Write(DeviceIndex: LongWord; Address, Reg, Value: Byte): LongWord; cdecl;
function DIO_SPI_Read(DeviceIndex: LongWord; Address, Reg: Byte; pValue: PByte): LongWord; cdecl;

implementation

uses
  Math,
  AIOUSB_Constants,
  VRTable,
  theseUtils,
  driverLevel,
  u8254,
  ADC, FastADC;

function DIO_SPI_Write(DeviceIndex: LongWord; Address, Reg, Value: Byte): LongWord; cdecl;
begin
  Result := ValidateAndEnsureOpen(DeviceIndex);
  if Result <> ERROR_SUCCESS then Exit;

  with Dev[DeviceIndex] do begin
    if not bDIOSPI then begin
      Result := ERROR_BAD_TOKEN_TYPE;
      Exit;
    end;

    Result := GenericVendorWrite(DeviceIndex, AUR_DIO_SPI_WRITE, Value, Reg or (Address shl 8), 0, nil);
  end;
end;

function DIO_SPI_Read(DeviceIndex: LongWord; Address, Reg: Byte; pValue: PByte): LongWord; cdecl;
var
  ReadValue: Byte;
  L: LongWord;

begin
  Result := ValidateAndEnsureOpen(DeviceIndex);
  if Result <> ERROR_SUCCESS then Exit;

  with Dev[DeviceIndex] do begin
    if not bDIOSPI then begin
      Result := ERROR_BAD_TOKEN_TYPE;
      Exit;
    end;

    L := SizeOf(ReadValue);
    Result := GenericVendorRead(DeviceIndex, AUR_DIO_SPI_READ, 0, Reg or (Address shl 8), L, @ReadValue);

    if Result = ERROR_SUCCESS then pValue^ := ReadValue;
  end;
end;


function DIO_CSA_DoSync(DeviceIndex: LongWord; var BaseRateHz, DurAms, DurBms, DurCms: Double): LongWord; cdecl;
const
  RootSyncClock = 3000;

var
  Divisor, DurATicks, DurBTicks, DurCTicks: Integer;
  Value, Index: Word;

begin
  Result := ValidateAndEnsureOpen(DeviceIndex);
  if Result <> ERROR_SUCCESS then Exit;

  with Dev[DeviceIndex] do begin
    if not bSetCustomClocks then begin
      Result := ERROR_BAD_TOKEN_TYPE;
      Exit;
    end;

    if BaseRateHz < (RootSyncClock/255.5) then
      Divisor := 0
    else if BaseRateHz >= RootSyncClock then
      Divisor := 1
    else
      Divisor := Min($FF, Round(RootSyncClock / BaseRateHz))
    ;

    if Divisor = 0 then begin
      BaseRateHz := 0;
      DurATicks := 0; DurAms := 0;
      DurBTicks := 0; DurBms := 0;
      DurCTicks := 0; DurCms := 0;
    end
    else begin
      BaseRateHz := RootSyncClock / Divisor;
      DurATicks := Min($FF, Round(3 * DurAms)); DurAms := 1/3 * DurATicks;
      DurBTicks := Min($FF, Round(3 * DurBms)); DurBms := 1/3 * DurBTicks;
      DurCTicks := Min($FF, Round(3 * DurCms)); DurCms := 1/3 * DurCTicks;
    end;
    Value := Divisor or (DurATicks shl 8);
    Index := DurBTicks or (DurCTicks shl 8);

    Result := GenericVendorWrite(DeviceIndex, AUR_SET_CUSTOM_CLOCKS, Value, Index, 0, nil);
  end;
end;

function DIO_CSA_DebounceSet(DeviceIndex, DebounceCounts: LongWord): LongWord; cdecl;
begin
  try
    Result := ValidateAndEnsureOpen(DeviceIndex);
    if Result <> ERROR_SUCCESS then Exit;

    with Dev[DeviceIndex] do begin
      if (DIOBytes = 0) or not bDIODebounce then begin
        Result := ERROR_BAD_TOKEN_TYPE;
        Exit;
      end;

      Result := GenericVendorWrite(DeviceIndex, AUR_DIO_ADVANCED, DebounceCounts, $100, 0, nil);
    end;
  except
    Result := ERROR_INTERNAL_ERROR;
  end;
end;

function DIO_CSA_DebounceReadAll(DeviceIndex: LongWord; pData: Pointer): LongWord; cdecl;
var
  L: LongWord;

begin
  try
    Result := ValidateAndEnsureOpen(DeviceIndex);
    if Result <> ERROR_SUCCESS then Exit;

    with Dev[DeviceIndex] do begin
      if (DIOBytes = 0) or not bDIODebounce then begin
        Result := ERROR_BAD_TOKEN_TYPE;
        Exit;
      end;

      L := DIOBytes * 2;
      Result := GenericVendorRead(DeviceIndex, AUR_DIO_ADVANCED, 0, $101, L, pData);
    end;
  except
    Result := ERROR_INTERNAL_ERROR;
  end;
end;

function DAC_CSA_SetRangeLimits(DeviceIndex: LongWord; pData: Pointer): LongWord; cdecl;
var
  InterBuf: AnsiString;

begin
  try
    SetString(InterBuf, PAnsiChar(pData), 32);
  except
    Result := ERROR_NOACCESS;
    Exit;
  end;
  InterBuf := #1 + InterBuf;
  Result := GenericVendorWrite(DeviceIndex, $A2, $1D00, 0, Length(InterBuf), PAnsiChar(InterBuf));
  InterBuf := '';
end;

function DAC_CSA_ClearRangeLimits(DeviceIndex: LongWord): LongWord; cdecl;
var
  InterBuf: AnsiString;

begin
  InterBuf := #0;
  Result := GenericVendorWrite(DeviceIndex, $A2, $1D00, 0, Length(InterBuf), PAnsiChar(InterBuf));
  InterBuf := '';
end;

function DIO_CSA_ReadAllString(DeviceIndex: LongWord; pData: Pointer; pDataBytes: PLongWord): LongWord; cdecl;
var
  L, LPer, BytesRead: LongWord;

begin
  try
    Result := Validate(DeviceIndex);
    if Result <> ERROR_SUCCESS then begin
      pDataBytes^ := 0;
      Exit;
    end;

    Result := EnsureOpen(DeviceIndex);
    if Result <> ERROR_SUCCESS then begin
      pDataBytes^ := 0;
      Exit;
    end;

    with Dev[DeviceIndex] do begin
      if DIOBytes = 0 then begin
        Result := ERROR_BAD_TOKEN_TYPE;
        pDataBytes^ := 0;
        Exit;
      end;

      LPer := DIOBytes;
      BytesRead := 0;
      L := pDataBytes^;
      Result := ERROR_SUCCESS;
      while L < LPer do begin
        Result := GenericVendorRead(DeviceIndex, AUR_DIO_READ, 0, 0, LPer, pData);
        if (Result <> ERROR_SUCCESS) or (LPer < DIOBytes) then Break;
        Inc(BytesRead, LPer);
        Dec(L, LPer);
        Inc(PByte(pData), LPer);
      end;
      pDataBytes^ := BytesRead;
    end;
  except
    Result := ERROR_INTERNAL_ERROR;
    pDataBytes^ := 0;
  end;
end;

function DACDIO_WriteAll(DeviceIndex: LongWord; pDACCounts: PWord; pDIOData: PByte): LongWord; cdecl;
var
  I: Integer;
  Content: AnsiString;

begin
  try

    Result := ValidateAndEnsureOpen(DeviceIndex);
    if Result <> ERROR_SUCCESS then Exit;

    with Dev[DeviceIndex] do begin
      if DIOBytes = 0 then begin
        Result := ERROR_BAD_TOKEN_TYPE;
        Exit;
      end;

      if ImmDACs = 0 then begin
        Result := ERROR_BAD_TOKEN_TYPE;
        Exit;
      end;

      SetLength(Content, 2 + DIOBytes + 2*ImmDACs);
      I := 1;

      Content[I] := AnsiChar(DIOBytes); //Length byte.
      Inc(I);

      Move(pDIOData^, Content[I], DIOBytes); //DIO.
      Inc(I, DIOBytes);

      Content[I] := AnsiChar(2*ImmDACs); //Length byte.
      Inc(I);

      Move(pDACCounts^, Content[I], 2*ImmDACs); //DACs.
      //Inc(BufI, 2*ImmDACs);

      Result := _GenericVendorWrite(DeviceIndex, AUR_DACDIO_WRITEALL, 0, 0, Length(Content), PAnsiChar(Content), Dev[DeviceIndex].bCyUSB);
    end;

  except
    Result := ERROR_INTERNAL_ERROR;
  end;
end;

function DIO_SetWatchdog16(DeviceIndex: LongWord; Delay, Data: Word): LongWord; cdecl; // deprecated
begin
  Result := ValidateAndEnsureOpen(DeviceIndex);
  if Result <> ERROR_SUCCESS then Exit;

  with Dev[DeviceIndex] do begin
    if DIOBytes = 0 then begin
      Result := ERROR_BAD_TOKEN_TYPE;
      Exit;
    end;

    case PID of
      $8010, $8011, $8012, $8014, $8015, $8016: ; //USB-IIRO-16 family
      else begin
        Result := ERROR_BAD_TOKEN_TYPE;
        Exit;
      end;
    end;

    Result := GenericVendorWrite(DeviceIndex, AUR_DIO_WDG16_DEPREC, Delay, Data, 0, nil);
  end;
end;

function ADC_InitFastITScanV(DeviceIndex: LongWord): LongWord; cdecl;
var
  I: Integer;
  L: LongWord;
  Dat: Byte;

begin
  try
    Result := ValidateAndEnsureOpen(DeviceIndex);
    if Result <> ERROR_SUCCESS then Exit;

    with Dev[DeviceIndex] do begin
      if not bADCBulk then begin
        Result := ERROR_BAD_TOKEN_TYPE;
        Exit;
      end;

      if ConfigBytes < 20 then begin
        Result := ERROR_BAD_TOKEN_TYPE;
        Exit;
      end;

      SetLength(FastITConfig, ConfigBytes);
      SetLength(FastITBakConfig, ConfigBytes);

      L := ConfigBytes;
      Result := ADC_GetConfig(DeviceIndex, @FastITBakConfig[0], L);
      if Result <> ERROR_SUCCESS then Exit;

      for I := 0 to 15 do FastITConfig[I] := FastITBakConfig[I]; //Use their range codes.
      //FastITConfig[$11] := $04 or (FastITBakConfig[$11] and $10); //Software-start scan, use their CTR0 EXT bit.
      FastITConfig[$11] := $05 or (FastITBakConfig[$11] and $10); //Timer scan, use their CTR0 EXT bit.
      FastITConfig[$13] := Max(3, FastITBakConfig[$13]); //Oversample at least +3.
      Dat := Min(64, ADCMUXChannels) - 1;
      FastITConfig[$12] := Dat shl 4;
      FastITConfig[$14] := Dat and $F0;
      Result := ADC_SetConfig(DeviceIndex, @FastITConfig[0], L);
      if Result <> ERROR_SUCCESS then begin
        ADC_SetConfig(DeviceIndex, @FastITBakConfig[0], L);
        Exit;
      end;
      Dat := $01;
      Result := GenericVendorWrite(DeviceIndex, $D4, $1E, 0, SizeOf(Dat), @Dat);
      if Result <> ERROR_SUCCESS then ADC_SetConfig(DeviceIndex, @FastITBakConfig[0], L);
    end;
  except
    Result := ERROR_INTERNAL_ERROR;
  end;
end;

function ADC_ResetFastITScanV(DeviceIndex: LongWord): LongWord; cdecl;
var
  L: LongWord;
  Dat: Byte;

begin
  try
    Result := ValidateAndEnsureOpen(DeviceIndex);
    if Result <> ERROR_SUCCESS then Exit;

    with Dev[DeviceIndex] do begin
      if not bADCBulk then begin
        Result := ERROR_BAD_TOKEN_TYPE;
        Exit;
      end;

      if ConfigBytes < 20 then begin
        Result := ERROR_BAD_TOKEN_TYPE;
        Exit;
      end;

      L := Length(FastITBakConfig);
      Result := ADC_SetConfig(DeviceIndex, @FastITBakConfig[0], L);
      if Result <> ERROR_SUCCESS then Exit;

      Dat := $00;
      Result := GenericVendorWrite(DeviceIndex, $D4, $1E, 0, SizeOf(Dat), @Dat);
    end;
  except
    Result := ERROR_INTERNAL_ERROR;
  end;
end;

function ADC_SetFastITScanVChannels(DeviceIndex, NewChannels: LongWord): LongWord; cdecl;
var
  L: LongWord;
  Dat: Byte;

begin
  try
    Result := ValidateAndEnsureOpen(DeviceIndex);
    if Result <> ERROR_SUCCESS then Exit;

    with Dev[DeviceIndex] do begin
      if not bADCBulk then begin
        Result := ERROR_BAD_TOKEN_TYPE;
        Exit;
      end;

      if ConfigBytes < 20 then begin
        Result := ERROR_BAD_TOKEN_TYPE;
        Exit;
      end;

      if Length(FastITConfig) < 20 then begin
        Result := ERROR_NOT_READY;
        Exit;
      end;

      L := Length(FastITConfig);
      Dat := Min(NewChannels, ADCMUXChannels) - 1;
      FastITConfig[$12] := Dat shl 4;
      FastITConfig[$14] := Dat and $F0;
      Result := ADC_SetConfig(DeviceIndex, @FastITConfig[0], L);
    end;
  except
    Result := ERROR_INTERNAL_ERROR;
  end;
end;

function ADC_GetFastITScanV(DeviceIndex: LongWord; pBuf: PDouble): LongWord; cdecl;
var
  ADBuf: TWordArray;
  I, Ch, J, Tot, Wt: Integer;
  V: Double;
  StartChannel, EndChannel, Channels, RangeCode: Byte;
  //Dummy: Word;
  BytesLeft: LongWord;

begin
  try
    Result := ValidateAndEnsureOpen(DeviceIndex);
    if Result <> ERROR_SUCCESS then Exit;

    with Dev[DeviceIndex] do begin
      if not bADCBulk then begin
        Result := ERROR_BAD_TOKEN_TYPE;
        Exit;
      end;

      if ConfigBytes < 20 then begin
        Result := ERROR_BAD_TOKEN_TYPE;
        Exit;
      end;

      //Determine channels.
      StartChannel := FastITConfig[$12] and $F;
      EndChannel := FastITConfig[$12] shr 4;
      if ConfigBytes >= 21 then begin
        StartChannel := StartChannel or ((FastITConfig[20] and $F) shl 4);
        EndChannel := EndChannel or (FastITConfig[20] and $F0);
      end;
      Channels := EndChannel - StartChannel + 1;

      //Set array to receive scan(=channels * (1+oversample)).
      SetLength(ADBuf, Channels * (1 + FastITConfig[$13]));

      //Begin get-data.
      Result := ADC_BulkAcquire(DeviceIndex, Length(ADBuf) * SizeOf(ADBuf[0]), @ADBuf[0]);
      if Result <> ERROR_SUCCESS then Exit;
      CTR_8254Mode(DeviceIndex, 0, 2, 0);
      CTR_8254Mode(DeviceIndex, 0, 2, 1);
      //ADC_GetImmediate(DeviceIndex, 0, @Dummy);

      //Complete get-data.
      if WaitForSingleObject(ADCWorker.Handle, 1000) <> WAIT_OBJECT_0 then begin
        Result := ERROR_TIMEOUT;
        Exit;
      end;
      Result := ADC_BulkPoll(DeviceIndex, BytesLeft);
      if Result <> ERROR_SUCCESS then Exit;
      if BytesLeft <> 0 then begin
        Result := ERROR_HANDLE_EOF;
        Exit;
      end;

      //Condition data; cull 75%(round down) of samples per channel from first block of channels, cull first from others, and average.
      I := 0;
      Inc(pBuf, StartChannel);
      for Ch := StartChannel to EndChannel do begin
        RangeCode := FastITConfig[Ch shr RangeShift];

        Tot := 0;
        Wt := 0;
        if (Ch shr RangeShift) = 0 then J := 4 else J := 1; //Discard first 4 samples from first 8 channels, and first 1 sample from other channels.
        J := Min(J, FastITConfig[$13]); //Except always keep at least one sample.
        for J := J to FastITConfig[$13] do begin
          Tot := Tot + ADBuf[I * (1 + FastITConfig[$13]) + J];
          Inc(Wt);
        end;

        V := Tot / Wt * (1 / 65536);
        if (RangeCode and 1) <> 0 then V := V * 2 - 1;
        if (RangeCode and 2) = 0 then V := V * 2;
        if (RangeCode and 4) = 0 then V := V * 5;
        pBuf^ := V;

        Inc(pBuf);
        Inc(I);
      end;
    end;
  except
    Result := ERROR_INTERNAL_ERROR;
  end;
end;

function ADC_GetITScanV(DeviceIndex: LongWord; pBuf: PDouble): LongWord; cdecl;
begin
  try
    Result := ADC_InitFastITScanV(DeviceIndex);
    if Result <> ERROR_SUCCESS then Exit;
    Result := ADC_GetFastITScanV(DeviceIndex, pBuf);
    if Result <> ERROR_SUCCESS then begin
      ADC_ResetFastITScanV(DeviceIndex);
      Exit;
    end;
    Result := ADC_ResetFastITScanV(DeviceIndex);
  except
    Result := ERROR_INTERNAL_ERROR;
  end;
end;


// J2H New Code added 20130401 for Jeff Smith: Convert "Fast IT" code to CSA generic

function ADC_CSA_InitFastScanV(DeviceIndex: LongWord): LongWord; cdecl;
var
  I: Integer;
  L: LongWord;

begin
  try
    Result := ValidateAndEnsureOpen(DeviceIndex);
    if Result <> ERROR_SUCCESS then Exit;

    with Dev[DeviceIndex] do begin
      if not bADCBulk then begin
        Result := ERROR_BAD_TOKEN_TYPE;
        Exit;
      end;

      if ConfigBytes < 20 then begin
        Result := ERROR_BAD_TOKEN_TYPE;
        Exit;
      end;

      SetLength(CSA_FastConfig, ConfigBytes);
      SetLength(CSA_FastBakConfig, ConfigBytes);

      L := ConfigBytes;
      Result := ADC_GetConfig(DeviceIndex, @CSA_FastBakConfig[0], L);     // backup existing config
      if Result <> ERROR_SUCCESS then Exit;

      for I := 0 to 15 do CSA_FastConfig[I] := CSA_FastBakConfig[I]; //Use their range codes.
      CSA_FastConfig[$11] := $05 or (CSA_FastBakConfig[$11] and $10); //Timer scan, use their CTR0 EXT bit.
      CSA_FastConfig[$13] := Max(1, CSA_FastBakConfig[$13]); //Oversample at least +1.

      //Use their start & end channel.
      CSA_FastConfig[$12] := CSA_FastBakConfig[$12];
      if ConfigBytes > $14 then
        CSA_FastConfig[$14] := CSA_FastBakConfig[$14]
      ;

      Result := ADC_SetConfig(DeviceIndex, @CSA_FastConfig[0], L);
      if Result <> ERROR_SUCCESS then begin
        ADC_SetConfig(DeviceIndex, @CSA_FastBakConfig[0], L);
        Exit;
      end;

      if Result <> ERROR_SUCCESS then ADC_SetConfig(DeviceIndex, @CSA_FastBakConfig[0], L);
    end;
  except
    Result := ERROR_INTERNAL_ERROR;
  end;
end;


function ADC_CSA_ResetFastScanV(DeviceIndex: LongWord): LongWord; cdecl;
var
  L: LongWord;

begin
  try
    Result := ValidateAndEnsureOpen(DeviceIndex);
    if Result <> ERROR_SUCCESS then Exit;

    with Dev[DeviceIndex] do begin
      if not bADCBulk then begin
        Result := ERROR_BAD_TOKEN_TYPE;
        Exit;
      end;

      if ConfigBytes < 20 then begin
        Result := ERROR_BAD_TOKEN_TYPE;
        Exit;
      end;

      L := Length(CSA_FastBakConfig);
      Result := ADC_SetConfig(DeviceIndex, @CSA_FastBakConfig[0], L);
      if Result <> ERROR_SUCCESS then Exit;
    end;
  except
    Result := ERROR_INTERNAL_ERROR;
  end;
end;



function ADC_CSA_GetFastScanV(DeviceIndex: LongWord; pBuf: PDouble): LongWord; cdecl;
var
  ADBuf: TWordArray;
  I, Ch, J, Tot, Wt: Integer;
  V: Double;
  StartChannel, EndChannel, Channels, RangeCode: Byte;
  //Dummy: Word;
  BytesLeft: LongWord;

begin
  try
    Result := ValidateAndEnsureOpen(DeviceIndex);
    if Result <> ERROR_SUCCESS then Exit;

    with Dev[DeviceIndex] do begin
      if not bADCBulk then begin
        Result := ERROR_BAD_TOKEN_TYPE;
        Exit;
      end;

      if ConfigBytes < 20 then begin
        Result := ERROR_BAD_TOKEN_TYPE;
        Exit;
      end;

      //Determine channels.
      StartChannel := CSA_FastConfig[$12] and $F;
      EndChannel := CSA_FastConfig[$12] shr 4;
      if ConfigBytes >= 21 then begin
        StartChannel := StartChannel or ((CSA_FastConfig[20] and $F) shl 4);
        EndChannel := EndChannel or (CSA_FastConfig[20] and $F0);
      end;
      Channels := EndChannel - StartChannel + 1;

      //Set array to receive scan(=channels * (1+oversample)).
      SetLength(ADBuf, Channels * (1 + CSA_FastConfig[$13]));

      //Begin get-data.
      Result := ADC_BulkAcquire(DeviceIndex, Length(ADBuf) * SizeOf(ADBuf[0]), @ADBuf[0]);
      if Result <> ERROR_SUCCESS then Exit;
      //ADC_GetImmediate(DeviceIndex, 0, @Dummy);

      //Complete get-data.
      if WaitForSingleObject(ADCWorker.Handle, 1000) <> WAIT_OBJECT_0 then begin
        Result := ERROR_TIMEOUT;
        Exit;
      end;
      Result := ADC_BulkPoll(DeviceIndex, BytesLeft);
      if Result <> ERROR_SUCCESS then Exit;
      if BytesLeft <> 0 then begin
        Result := ERROR_HANDLE_EOF;
        Exit;
      end;

      //Condition data; cull first sample per channel, and average.
      I := 0;
      Inc(pBuf, StartChannel);
      for Ch := StartChannel to EndChannel do begin
        RangeCode := CSA_FastConfig[Ch shr RangeShift];

        Tot := 0;
        Wt := 0;
        J := Min(1, CSA_FastConfig[$13]); //Discard first sample, except always keep at least one sample.
        for J := J to CSA_FastConfig[$13] do begin
          Tot := Tot + ADBuf[I * (1 + CSA_FastConfig[$13]) + J];
          Inc(Wt);
        end;

        V := Tot / Wt * (1 / 65536);
        if (RangeCode and 1) <> 0 then V := V * 2 - 1;
        if (RangeCode and 2) = 0 then V := V * 2;
        if (RangeCode and 4) = 0 then V := V * 5;
        pBuf^ := V;

        Inc(pBuf);
        Inc(I);
      end;
    end;
  except
    Result := ERROR_INTERNAL_ERROR;
  end;
end;



end.
