unit ADC;

interface

uses
  Windows;

function ADC_GetConfig(DeviceIndex: LongWord; pConfigBuf: Pointer; var ConfigBufSize: LongWord): LongWord; cdecl;
function ADC_SetConfig(DeviceIndex: LongWord; pConfigBuf: Pointer; var ConfigBufSize: LongWord): LongWord; cdecl;
function ADC_RangeAll(DeviceIndex: LongWord; pGainCodes: PByte; bDifferential: LongBool): LongWord; cdecl;
function ADC_Range1(DeviceIndex, ADChannel: LongWord; GainCode: Byte; bDifferential: LongBool): LongWord; cdecl;
function ADC_ADMode(DeviceIndex: LongWord; TriggerMode, CalMode: Byte): LongWord; cdecl;
function ADC_SetScanLimits(DeviceIndex, StartChannel, EndChannel: LongWord): LongWord; cdecl;
function ADC_SetOversample(DeviceIndex: LongWord; Oversample: Byte): LongWord; cdecl;
function ADC_QueryCal(DeviceIndex: LongWord): LongWord; cdecl;
function ADC_SetCal(DeviceIndex: LongWord; CalFileName: PAnsiChar): LongWord; cdecl;
function ADC_SetCalAndSave(DeviceIndex: LongWord; CalFileName, SaveCalFileName: PAnsiChar): LongWord; cdecl;
function ADC_Initialize(DeviceIndex: LongWord; pConfigBuf: Pointer; var ConfigBufSize: LongWord; CalFileName: PAnsiChar): LongWord; cdecl;
function ADC_Start(DeviceIndex: LongWord): LongWord; cdecl;
function ADC_Stop(DeviceIndex: LongWord): LongWord; cdecl;
function ADC_GetImmediate(DeviceIndex, Channel: LongWord; pBuf: PWord): LongWord; cdecl;
function ADC_GetScan(DeviceIndex: LongWord; pBuf: PWord): LongWord; cdecl;
function ADC_GetScanV(DeviceIndex: LongWord; pBuf: PDouble): LongWord; cdecl;
function ADC_GetTrigScan(DeviceIndex: LongWord; pBuf: PWord; TimeoutMS: LongInt): LongWord; cdecl;
function ADC_GetTrigScanV(DeviceIndex: LongWord; pBuf: PDouble; TimeoutMS: LongInt): LongWord; cdecl;
function ADC_GetChannelV(DeviceIndex, ChannelIndex: LongWord; pBuf: PDouble): LongWord; cdecl;
function ADC_GetCalRefV(DeviceIndex, CalRefIndex: LongWord; var pRef: Double): LongWord; cdecl;

implementation

uses
{$IFDEF VER130}  // Delphi 5
  SysUtils, Classes,
{$ELSE}         // anything other than Delphi 5, presumably XE2 or XE3
  System.SysUtils,
  System.Classes,
{$ENDIF}
  Math,
  DeviceLevel,
  VRTable,
  theseUtils,
  AIOUSB_Constants,
  AIOUSB_Types,
  u8254,
  FastADC,
  ContBufThreads
  ;

function ADC_GetConfig(DeviceIndex: LongWord; pConfigBuf: Pointer; var ConfigBufSize: LongWord): LongWord; cdecl;
var
  Content: AnsiString;
  L: LongWord;

begin
  Result := ValidateAndEnsureOpen(DeviceIndex);
  if Result <> ERROR_SUCCESS then Exit;

  with Dev[DeviceIndex] do begin
    if ConfigBytes = 0 then begin
      Result := ERROR_BAD_TOKEN_TYPE;
      Exit;
    end;

    if ConfigBufSize < ConfigBytes then begin
      Result := ERROR_INSUFFICIENT_BUFFER;
      ConfigBufSize := ConfigBytes;
      Exit;
    end;

    if pConfigBuf = nil then begin
      Result := ERROR_INVALID_PARAMETER;
      Exit;
    end;

    SetString(Content, PAnsiChar(pConfigBuf), ConfigBytes); //Probe buffer

    L := ConfigBytes;
    Result := GenericVendorRead(DeviceIndex, AUR_ADC_GET_CONFIG, 0, 0, L, pConfigBuf);
    if Result = ERROR_SUCCESS then ConfigBufSize := ConfigBytes;
  end;
end;

function ADC_SetConfig(DeviceIndex: LongWord; pConfigBuf: Pointer; var ConfigBufSize: LongWord): LongWord; cdecl;
var
  Content: AnsiString;

begin
  Result := ValidateAndEnsureOpen(DeviceIndex);
  if Result <> ERROR_SUCCESS then Exit;

  with Dev[DeviceIndex] do begin
    if ConfigBytes = 0 then begin
      Result := ERROR_BAD_TOKEN_TYPE;
      Exit;
    end;

    if ConfigBufSize < ConfigBytes then begin
      Result := ERROR_INSUFFICIENT_BUFFER;
      ConfigBufSize := ConfigBytes;
      Exit;
    end;

    if pConfigBuf = nil then begin
      Result := ERROR_INVALID_PARAMETER;
      Exit;
    end;

    SetString(Content, PAnsiChar(pConfigBuf), ConfigBytes); //Probe buffer

    Result := GenericVendorWrite(DeviceIndex, AUR_ADC_SET_CONFIG, 0, 0, ConfigBytes, pConfigBuf);

    if Result = ERROR_SUCCESS then ConfigBufSize := ConfigBytes;
  end;
end;

function ADC_RangeAll(DeviceIndex: LongWord; pGainCodes: PByte; bDifferential: LongBool): LongWord; cdecl;
var
  ConfigStr: AnsiString;
  L: LongWord;
  I: Integer;
  pRC: PByte;

begin
  Result := ValidateAndEnsureOpen(DeviceIndex);
  if Result <> ERROR_SUCCESS then Exit;

  with Dev[DeviceIndex] do begin
    if (ADCChannels = 0) or (not bADCBulk) then begin
      Result := ERROR_BAD_TOKEN_TYPE;
      Exit;
    end;

    if pGainCodes = nil then begin
      Result := ERROR_INVALID_PARAMETER;
      Exit;
    end;

    pRC := pGainCodes;
    for I := 1 to ADCChannels do begin
      if (pRC^ and $F8) <> 0 then begin
        Result := ERROR_INVALID_DATA;
        Exit;
      end;
      Inc(pRC);
    end;

    L := ConfigBytes;
    SetString(ConfigStr, nil, L);

    Result := GenericVendorRead(DeviceIndex, AUR_ADC_GET_CONFIG, 0, 0, L, PAnsiChar(ConfigStr));
    if Result <> ERROR_SUCCESS then Exit;

    pRC := pGainCodes;
    for I := 1 to ADCChannels do begin
      ConfigStr[I] := AnsiChar(pRC^ or (Ord(bDifferential) shl 3));
      Inc(pRC);
    end;

    Result := GenericVendorWrite(DeviceIndex, AUR_ADC_SET_CONFIG, 0, 0, Length(ConfigStr), PAnsiChar(ConfigStr));
  end;
end;

function ADC_Range1(DeviceIndex, ADChannel: LongWord; GainCode: Byte; bDifferential: LongBool): LongWord; cdecl;
var
  ConfigStr: AnsiString;
  L: LongWord;

begin
  Result := ValidateAndEnsureOpen(DeviceIndex);
  if Result <> ERROR_SUCCESS then Exit;

  with Dev[DeviceIndex] do begin
    if not bADCBulk then begin
      Result := ERROR_BAD_TOKEN_TYPE;
      Exit;
    end;

    if ADChannel >= ADCChannels then begin
      Result := ERROR_INVALID_ADDRESS;
      Exit;
    end;

    if (GainCode and $F8) <> 0 then begin
      Result := ERROR_INVALID_PARAMETER;
      Exit;
    end;

    L := ConfigBytes;
    SetString(ConfigStr, nil, L);

    Result := GenericVendorRead(DeviceIndex, AUR_ADC_GET_CONFIG, 0, 0, L, PAnsiChar(ConfigStr));
    if Result <> ERROR_SUCCESS then Exit;

    ConfigStr[1 + (ADChannel mod ADCChannels)] := AnsiChar(GainCode or (Ord(bDifferential) shl 3));

    Result := GenericVendorWrite(DeviceIndex, AUR_ADC_SET_CONFIG, 0, 0, Length(ConfigStr), PAnsiChar(ConfigStr));
  end;
end;

function ADC_ADMode(DeviceIndex: LongWord; TriggerMode, CalMode: Byte): LongWord; cdecl;
var
  ConfigStr: AnsiString;
  L: LongWord;

begin
  Result := ValidateAndEnsureOpen(DeviceIndex);
  if Result <> ERROR_SUCCESS then Exit;

  with Dev[DeviceIndex] do begin
    if not bADCBulk then begin
      Result := ERROR_BAD_TOKEN_TYPE;
      Exit;
    end;

    if not CalMode in [$00, $01, $03, $05, $07] then begin
      Result := ERROR_INVALID_PARAMETER;
      Exit;
    end;

    L := ConfigBytes;
    SetString(ConfigStr, nil, L);

    Result := GenericVendorRead(DeviceIndex, AUR_ADC_GET_CONFIG, 0, 0, L, PAnsiChar(ConfigStr));
    if Result <> ERROR_SUCCESS then Exit;

    ConfigStr[1 + $10] := AnsiChar(CalMode);
    ConfigStr[1 + $11] := AnsiChar(TriggerMode);

    Result := GenericVendorWrite(DeviceIndex, AUR_ADC_SET_CONFIG, 0, 0, Length(ConfigStr), PAnsiChar(ConfigStr));
  end;
end;

function ADC_SetScanLimits(DeviceIndex, StartChannel, EndChannel: LongWord): LongWord; cdecl;
var
  ConfigStr: AnsiString;
  L: LongWord;

begin
  Result := ValidateAndEnsureOpen(DeviceIndex);
  if Result <> ERROR_SUCCESS then Exit;

  with Dev[DeviceIndex] do begin
    if not bADCBulk then begin
      Result := ERROR_BAD_TOKEN_TYPE;
      Exit;
    end;

    if (StartChannel > EndChannel) or (EndChannel > ADCMUXChannels) then begin
      Result := ERROR_INVALID_PARAMETER;
      Exit;
    end;

    L := ConfigBytes;
    SetString(ConfigStr, nil, L);

    Result := GenericVendorRead(DeviceIndex, AUR_ADC_GET_CONFIG, 0, 0, L, PAnsiChar(ConfigStr));
    if Result <> ERROR_SUCCESS then Exit;

    if ConfigBytes < $15 then
      ConfigStr[1 + $12] := AnsiChar(StartChannel or (EndChannel shl 4))
    else begin
      ConfigStr[1 + $12] := AnsiChar((StartChannel and $0F) or (EndChannel shl 4));
      ConfigStr[1 + $14] := AnsiChar((StartChannel shr 4) or (EndChannel and $F0));
    end;

    Result := GenericVendorWrite(DeviceIndex, AUR_ADC_SET_CONFIG, 0, 0, Length(ConfigStr), PAnsiChar(ConfigStr));
  end;
end;


function ADC_SetOversample(DeviceIndex: LongWord; Oversample: Byte): LongWord; cdecl;
var
  ConfigStr: AnsiString;
  L: LongWord;

begin
  Result := ValidateAndEnsureOpen(DeviceIndex);
  if Result <> ERROR_SUCCESS then Exit;

  with Dev[DeviceIndex] do begin
    if not bADCBulk then begin
      Result := ERROR_BAD_TOKEN_TYPE;
      Exit;
    end;

    L := ConfigBytes;
    SetString(ConfigStr, nil, L);

    Result := GenericVendorRead(DeviceIndex, AUR_ADC_GET_CONFIG, 0, 0, L, PAnsiChar(ConfigStr));
    if Result <> ERROR_SUCCESS then Exit;

    ConfigStr[1 + $13] := AnsiChar(Oversample);

    Result := GenericVendorWrite(DeviceIndex, AUR_ADC_SET_CONFIG, 0, 0, Length(ConfigStr), PAnsiChar(ConfigStr));
  end;
end;

function ADC_QueryCal(DeviceIndex: LongWord): LongWord; cdecl;
var
  L: LongWord;
  DataByte: Byte;
  Status: LongWord;

begin
  Result := ValidateAndEnsureOpen(DeviceIndex);
  if Result <> ERROR_SUCCESS then Exit;

  with Dev[DeviceIndex] do begin
    if not bADCBulk then begin
      Result := ERROR_BAD_TOKEN_TYPE;
      Exit;
    end;

    DataByte := $FF;
    L := SizeOf(DataByte);
    Status := GenericVendorRead(DeviceIndex, AUR_PROBE_CALFEATURE, 0, 0, L, @DataByte);
    if Status <> ERROR_SUCCESS then begin
      Result := Status;
      Exit;
    end;
    if (L <> SizeOf(DataByte)) or (DataByte <> $BB) then begin
      Result := ERROR_NOT_SUPPORTED;
      Exit;
    end;

    Result := ERROR_SUCCESS;
  end;
end;

function ADC_SetCal(DeviceIndex: LongWord; CalFileName: PAnsiChar): LongWord; cdecl;
begin
  Result := ADC_SetCalAndSave(DeviceIndex, CalFileName, nil);
end;

function ADC_SetCalAndSave(DeviceIndex: LongWord; CalFileName, SaveCalFileName: PAnsiChar): LongWord; cdecl;
const
  ChunkSize = $400;

var
  CalTable: array of Word;
  oConfig, nConfig: array of Byte;
  Suc: Boolean;

  procedure DoLoadCalTable;
  var
    bFirstChunkDouble: Boolean;
    TargetPipe: LongWord;
    SRAMIndex: Integer;
    L, L2: LongWord;

  begin
    bFirstChunkDouble := True;
    TargetPipe := $02;
    SRAMIndex := 0;
    repeat
      L2 := (Length(CalTable) - SRAMIndex);
      if L2 > ChunkSize then L2 := ChunkSize;
      L := 0;
      Suc := AWU_GenericBulkOut(DeviceIndex, TargetPipe, @CalTable[SRAMIndex], SizeOf(CalTable[0]) * L2, L) = ERROR_SUCCESS;
      if not Suc then Break;

      GenericVendorWrite(DeviceIndex, $BB, SRAMIndex, L2, 0, nil);

      L := L div 2; //Entries
      L := L and $FF00; //Entries in complete packets
      Inc(SRAMIndex, L);
      if bFirstChunkDouble then begin
        bFirstChunkDouble := False;
        SRAMIndex := 0;
      end;
    until (SRAMIndex = $10000);
  end;

  function ConfigureAndAcquire: Double;
  var
    L: LongWord;
    I, Tot: Integer;
    Sample: Word;

  begin
    L := Length(nConfig);
    ADC_SetConfig(DeviceIndex, @nConfig[0], L);
    Tot := 0;
    for I := 1 to 256 do begin
      ADC_GetImmediate(DeviceIndex, 0, @Sample);
      Inc(Tot, Sample);
    end;
    Result := Tot * (1 / 256);
  end;

  function ConfigureAndBulkAcquire: Double;
  var
    L, BytesLeft, TargetPipe, BCData: LongWord;
    I, Tot: Integer;
    ADData: array of Word;
    pADBuf: PByte;

  begin
    L := Length(nConfig);
    ADC_SetConfig(DeviceIndex, @nConfig[0], L);
    SetLength(ADData, 256);

    //Inline bulk the data.
    TargetPipe := $86;
    BytesLeft := SizeOf(ADData[0]) * Length(ADData);
    pADBuf := @ADData[0];

    //GenericVendorWrite(DeviceIndex, AUR_SET_GPIF_MODE, 0, $0, 0, nil);
    BCData := $00000005;
    GenericVendorWrite(DeviceIndex, $BC, 0, Length(ADData), SizeOf(BCData), @BCData);
    ADC_GetImmediate(DeviceIndex, 0, @ADData[0]);
    repeat
      L := 0;
      Result := AWU_GenericBulkIn(DeviceIndex, TargetPipe, pADBuf, BytesLeft, L);
      if Result <> ERROR_SUCCESS then Break;
      Dec(BytesLeft, L);
      Inc(pADBuf, L);
    until BytesLeft <= 0;

    //Average it up.
    Tot := 0;
    for I := 0 to High(ADData) do Inc(Tot, ADData[I]);
    Result := Tot / Length(ADData);
  end;

  function GetHiRef: Word;
  var
    Status, DataSize: LongWord;

  begin
    DataSize := 2;
    Status := GenericVendorRead(DeviceIndex, $A2, $1DF2, 0, DataSize, @Result);
    if Status <> ERROR_SUCCESS then
      Result := 0
    else if DataSize <> 2 then
      Result := 0
    else if Result = $FFFF then
      Result := 0
    ;
  end;

  function ReadCalTableFromFile(hFil: THandle): LongWord;
  var
    L: LongWord;

  begin
    L := 0;
    Suc := ReadFile(hFil, CalTable[0], SizeOf(CalTable[0]) * Length(CalTable), L, nil);
    Result := GetLastError;
    if (Result = ERROR_SUCCESS) and not Suc then Result := ERROR_GEN_FAILURE;
    if (Result = ERROR_SUCCESS) and (LongInt(L) < SizeOf(CalTable[0]) * Length(CalTable)) then Result := ERROR_HANDLE_EOF;
  end;

  function WriteCalTableToFile(hFil: THandle): LongWord;
  var
    L: LongWord;

  begin
    if hFil = 0 then begin //Successfully does nothing.
      Result := ERROR_SUCCESS;
      Exit;
    end;

    L := 0;
    Suc := WriteFile(hFil, CalTable[0], SizeOf(CalTable[0]) * Length(CalTable), L, nil);
    Result := GetLastError;
    if (Result = ERROR_SUCCESS) and not Suc then Result := ERROR_GEN_FAILURE;
    if (Result = ERROR_SUCCESS) and (LongInt(L) < SizeOf(CalTable[0]) * Length(CalTable)) then Result := ERROR_HANDLE_EOF;
  end;

const
  LoRefRef = 0      * 6553.6;
  HiRefRef = 9.9339 * 6553.6;

var
  I, J, K: Integer;
  L: LongWord;
  hFil, hSaveFil: THandle;
  F,
  LoRef,  HiRef,  dRef,
  LoRead, HiRead, dRead: Double;
  DataByte: Byte;
  Status: LongWord;

begin
  Result := ValidateAndEnsureOpen(DeviceIndex);
  if Result <> ERROR_SUCCESS then Exit;

  with Dev[DeviceIndex] do begin
    if not bADCBulk then begin
      Result := ERROR_BAD_TOKEN_TYPE;
      Exit;
    end;

    DataByte := $FF;
    L := SizeOf(DataByte);
    Status := GenericVendorRead(DeviceIndex, AUR_PROBE_CALFEATURE, 0, 0, L, @DataByte);
    if Status <> ERROR_SUCCESS then begin
      Result := Status;
      Exit;
    end;
    if (L <> SizeOf(DataByte)) or (DataByte <> $BB) then begin
      Result := ERROR_NOT_SUPPORTED;
      Exit;
    end;

    if SaveCalFileName = nil then
      hSaveFil := 0
    else begin
      hSaveFil := CreateFileA(SaveCalFileName, GENERIC_READ or GENERIC_WRITE, FILE_SHARE_READ or FILE_SHARE_WRITE, nil, CREATE_ALWAYS, 0, 0);
      if (hSaveFil = 0) or (hSaveFil = INVALID_HANDLE_VALUE) then begin
        Result := GetLastError;
        if Result = ERROR_SUCCESS then Result := ERROR_GEN_FAILURE;
        Exit;
      end;
    end;

    L := ConfigBytes;
    SetLength(oConfig, L);
    SetLength(nConfig, L);
    FillChar(oConfig[0], L, $00);
    FillChar(nConfig[0], L, $00);
    ADC_GetConfig(DeviceIndex, @oConfig[0], L);

    try
      if CalFileName = ':AUTO:' then begin

        LoRef := LoRefRef;
        HiRef := GetHiRef;
        if HiRef = 0 then HiRef := HiRefRef;
        dRef := HiRef - LoRef;

        nConfig[$10] := $05; //Select bip low ref, to select bip cal bank(on banked boards).
        nConfig[$00] := $01; //Select 10V range for channel 0.
        for K := 0 to 1 do begin
          ADC_SetConfig(DeviceIndex, @nConfig[0], L);

          //Load 1-to-1 to cal against.
          SetLength(CalTable, $10000);
          for I := 0 to High(CalTable) do CalTable[I] := I;
          DoLoadCalTable;

          nConfig[$11] := $04; //scan software-start
          nConfig[$12] := $00; //select 1 channel
          nConfig[$13] := $FF; //+255 oversample per channel, for 256 samples per scan

          nConfig[$10] := nConfig[$10] or $02; //cal high ref
          HiRead := ConfigureAndBulkAcquire; //$FE3F;
          if HiRead < $F000 then begin
            Result := ERROR_INVALID_DATA;
            Exit;
          end;
          Sleep(10); //Cargo cult.

          nConfig[$10] := nConfig[$10] and not $02; //cal low ref
          LoRead := ConfigureAndBulkAcquire; //$0042;
          if K = 0 then begin
            if (LoRead < $7F00) or (LoRead > $80FF) then begin
              Result := ERROR_INVALID_DATA;
              Exit;
            end;
          end
          else begin
            if LoRead > $00FF then begin
              Result := ERROR_INVALID_DATA;
              Exit;
            end;
          end;
          Sleep(10); //Cargo cult.

          dRead := HiRead - LoRead;

          SetLength(CalTable, $10000);
          for I := 0 to High(CalTable) do begin
            F := (I - LoRead) / dRead;
            F := LoRef + F * dRef;
            if K = 0 then F := 0.5 * (F + $10000);
            J := Round(F);
            if J <= 0 then
              J := 0
            else if J >= $FFFF then
              J := $FFFF
            ;
            CalTable[I] := J;
          end;

          DoLoadCalTable;

          Result := WriteCalTableToFile(hSaveFil);
          if Result <> ERROR_SUCCESS then Exit;

          nConfig[$10] := $01; //Select unip low ref, to select unip cal bank(on banked boards).
          nConfig[$00] := $00; //Select 0-10V range for channel 0.
        end;

      end
      else if (CalFileName = ':NORM:') then begin

        LoRef := LoRefRef;
        HiRef := GetHiRef;
        if HiRef = 0 then HiRef := HiRefRef;
        dRef := HiRef - LoRef;

        LoRead := $0042;
        HiRead := $FE3F;
        dRead := HiRead - LoRead;

        SetLength(CalTable, $10000);
        for I := 0 to High(CalTable) do begin
          F := (I - LoRead) / dRead;
          F := LoRef + F * dRef;
          J := Round(F);
          if J <= 0 then
            J := 0
          else if J >= $FFFF then
            J := $FFFF
          ;
          CalTable[I] := J;
        end;

        nConfig[$10] := $05; //Select bip low ref, to select bip cal bank(on banked boards).
        for K := 0 to 1 do begin
          ADC_SetConfig(DeviceIndex, @nConfig[0], L);

          DoLoadCalTable;

          Result := WriteCalTableToFile(hSaveFil);
          if Result <> ERROR_SUCCESS then Exit;

          nConfig[$10] := $01; //Select unip low ref, to select unip cal bank(on banked boards).
        end;

      end
      else if (CalFileName = ':1TO1:') or (CalFileName = ':NONE:') then begin

        SetLength(CalTable, $10000);
        for I := 0 to High(CalTable) do CalTable[I] := I;

        nConfig[$10] := $05; //Select bip low ref, to select bip cal bank(on banked boards).
        for K := 0 to 1 do begin
          ADC_SetConfig(DeviceIndex, @nConfig[0], L);

          DoLoadCalTable;

          Result := WriteCalTableToFile(hSaveFil);
          if Result <> ERROR_SUCCESS then Exit;

          nConfig[$10] := $01; //Select unip low ref, to select unip cal bank(on banked boards).
        end;

      end
      else if Copy(CalFileName, 1, 1) = ':' then begin

        Result := ERROR_MOD_NOT_FOUND;
        Exit;

      end
      else begin //Not one of the macros, so treat it as a filename.

        hFil := CreateFileA(CalFileName, GENERIC_READ, FILE_SHARE_READ or FILE_SHARE_WRITE, nil, OPEN_EXISTING, 0, 0);
        if (hFil = 0) or (hFil = INVALID_HANDLE_VALUE) then begin
          Result := GetLastError;
          if Result = ERROR_SUCCESS then Result := ERROR_GEN_FAILURE;
          Exit;
        end;
        SetLength(CalTable, $10000);

        I := GetFileSize(hFil, @L);
        Result := GetLastError;
        if (LongWord(I) = INVALID_FILE_SIZE) and (Result <> ERROR_SUCCESS) then begin
          CloseHandle(hFil);
          Exit;
        end;

        if (L <> 0) or (I >= $20000) then begin
          //It's big enough to be a "double" binary cal file.
          nConfig[$10] := $05; //Select bip low ref, to select bip cal bank(on banked boards).
          nConfig[$00] := $01; //Select 10V range for channel 0.
          for K := 0 to 1 do begin
            ADC_SetConfig(DeviceIndex, @nConfig[0], L);

            Result := ReadCalTableFromFile(hFil);
            if Result <> ERROR_SUCCESS then begin
              CloseHandle(hFil);
              Exit;
            end;

            DoLoadCalTable;

            Result := WriteCalTableToFile(hSaveFil);
            if Result <> ERROR_SUCCESS then begin
              CloseHandle(hFil);
              Exit;
            end;

            nConfig[$10] := $01; //Select unip low ref, to select unip cal bank(on banked boards).
            nConfig[$00] := $00; //Select 0-10V range for channel 0.
          end;
          CloseHandle(hFil);
        end
        else if I >= $10000 then begin
          //It's only a "single" binary cal file.
          Result := ReadCalTableFromFile(hFil);
          CloseHandle(hFil);
          if Result <> ERROR_SUCCESS then Exit;

          DoLoadCalTable;

          Result := WriteCalTableToFile(hSaveFil);
          if Result <> ERROR_SUCCESS then Exit;
        end
        else begin
          Result := ERROR_HANDLE_EOF;
          CloseHandle(hFil);
          Exit;
        end;

      end;
    finally
      L := ConfigBytes;
      ADC_SetConfig(DeviceIndex, @oConfig[0], L);
      SetLength(oConfig, 0);
      SetLength(nConfig, 0);

      SetLength(CalTable, 0);

      if hSaveFil <> 0 then CloseHandle(hSaveFil);
    end;

    if Suc then
      Result := ERROR_SUCCESS
    else
      Result := GetMappedFailure(DeviceIndex)
    ;
  end;
end;

function ADC_Initialize(DeviceIndex: LongWord; pConfigBuf: Pointer; var ConfigBufSize: LongWord; CalFileName: PAnsiChar): LongWord; cdecl;
begin
  Result := ERROR_SUCCESS;
  if (pConfigBuf <> nil) and (ConfigBufSize <> 0) then begin
    Result := ADC_SetConfig(DeviceIndex, pConfigBuf, ConfigBufSize);
    if Result <> ERROR_SUCCESS then Exit;
  end;

  if CalFileName <> nil then begin
    Result := ADC_SetCal(DeviceIndex, CalFileName);
    //if Result <> ERROR_SUCCESS then Exit;
  end;
end;

function ADC_Start(DeviceIndex: LongWord): LongWord; cdecl;
begin
  Result := ERROR_NOT_SUPPORTED;
end;

function ADC_Stop(DeviceIndex: LongWord): LongWord; cdecl;
begin
  Result := ERROR_NOT_SUPPORTED;
end;

function ADC_GetImmediate(DeviceIndex, Channel: LongWord; pBuf: PWord): LongWord; cdecl;
var
  L: LongWord;
  oADCWorker: TADCWorker;

begin
  Result := ValidateAndEnsureOpen(DeviceIndex);
  if Result <> ERROR_SUCCESS then Exit;

  with Dev[DeviceIndex] do begin
    if not bADCBulk then begin
      Result := ERROR_BAD_TOKEN_TYPE;
      Exit;
    end;

    {
    if not bADCOpen then begin
      Result := ERROR_FILE_NOT_FOUND;
      Exit;
    end;
    }

    oADCWorker := TADCWorker(ADCWorker);
    if (oADCWorker <> nil) and (oADCWorker.BytesLeft <> 0) then begin
      Result := ERROR_TOO_MANY_OPEN_FILES;
      Exit;
    end;

    try
      pBuf^ := $0000; //Probe buffer
    except
      Result := ERROR_NOACCESS;
      Exit;
    end;

    L := 0;
    Result := GenericVendorWrite(DeviceIndex, AUR_ADC_IMMEDIATE, 0, Channel, L, nil);
  end;
end;

type
  TTimeoutTriggerThread = class(TStartableThread)
  public
    DI: Integer;
    TimeoutMS: LongInt;
    bTimeout: Boolean;
  protected
    procedure Execute; override;
  end;

procedure TTimeoutTriggerThread.Execute;
var
  StartTime, ThisTime, DeltaTime, TimeoutTime: Int64;
  L: LongWord;

begin
  bTimeout := False;
  QueryPerformanceFrequency(TimeoutTime);
  TimeoutTime := (TimeoutTime * TimeoutMS) div 1000;
  QueryPerformanceCounter(StartTime);
  repeat
    Sleep(1);
    if Terminated then Break;

    QueryPerformanceCounter(ThisTime);
    DeltaTime := ThisTime - StartTime;
    if DeltaTime > TimeoutTime then begin
      L := 0;
      GenericVendorWrite(DI, AUR_ADC_IMMEDIATE, 0, 0, L, nil);
      bTimeout := True;
      Break;
    end;
  until False;
end;

function ADC_GetScan_Inner(var DeviceIndex: LongWord; var Config: array of Byte; var ADBuf: TWordArray; var StartChannel, EndChannel: Byte; TimeoutMS: LongInt): LongWord;
var
  pADBuf: PByte;
  L, BytesLeft, TargetPipe, BCData: LongWord;
  Channels: Byte;
  I: Integer;
  TimeoutTriggerThread: TTimeoutTriggerThread;

begin
  Result := ValidateAndEnsureOpen(DeviceIndex);
  if Result <> ERROR_SUCCESS then Exit;

  if Dev[DeviceIndex].bADCBulk then begin
    L := SizeOf(Config);
    Result := ADC_GetConfig(DeviceIndex, @Config[0], L);
    if Result <> ERROR_SUCCESS then Exit;

    if TimeoutMS <> 0 then
      Config[$11] := $04 or Config[$11] //Turn scan on.
    else
      Config[$11] := $04 or Config[$11] and not $03 //Turn scan on, turn timer and external trigger off.
    ;
    StartChannel := Config[$12] and $F;
    EndChannel := Config[$12] shr 4;
    if L >= 21 then begin
      StartChannel := StartChannel or ((Config[20] and $F) shl 4);
      EndChannel := EndChannel or (Config[20] and $F0);
    end;
    Channels := EndChannel - StartChannel + 1;
    Config[$13] := Max(1, Config[$13]); //Oversample at least +1.
    if Channels * (1 + Config[$13]) > 1024 then Config[$13] := (1024 div Channels) - 1; //Don't take more than the board can buffer, since we aren't threading.
    SetLength(ADBuf, Channels * (1 + Config[$13]));
    Result := ADC_SetConfig(DeviceIndex, @Config[0], L);
    if Result <> ERROR_SUCCESS then Exit;

    TargetPipe := $86;
    BytesLeft := SizeOf(ADBuf[0]) * Length(ADBuf);
    pADBuf := @ADBuf[0];

    //GenericVendorWrite(DeviceIndex, AUR_SET_GPIF_MODE, 0, $0, 0, nil);
    BCData := $00000005;
    Result := GenericVendorWrite(DeviceIndex, $BC, 0, Length(ADBuf), SizeOf(BCData), @BCData);
    if Result <> ERROR_SUCCESS then Exit;
    if TimeoutMS <> 0 then begin
      //Spawn timeout-trigger thread.
      TimeoutTriggerThread := TTimeoutTriggerThread.Create(True);
      TimeoutTriggerThread.FreeOnTerminate := False;
      TimeoutTriggerThread.DI := DeviceIndex;
      TimeoutTriggerThread.TimeoutMS := TimeoutMS;
      TimeoutTriggerThread.Start;
    end
    else begin
      ADC_GetImmediate(DeviceIndex, 0, @ADBuf[0]);
      TimeoutTriggerThread := nil;
    end;
    repeat
      L := 0;
      Result := AWU_GenericBulkIn(DeviceIndex, TargetPipe, pADBuf, BytesLeft, L);
      if Result <> ERROR_SUCCESS then Break;
      Dec(BytesLeft, L);
      Inc(pADBuf, L);
    until BytesLeft <= 0;
    if TimeoutTriggerThread <> nil then begin
      //Tell timeout-trigger thread to not bother if it hasn't tripped yet.
      TimeoutTriggerThread.Terminate;
      //Report timeout if it did.
      if (Result = ERROR_SUCCESS) and TimeoutTriggerThread.bTimeout then Result := ERROR_TIMEOUT;
      //And clean up the thread.
      TimeoutTriggerThread.Free;
    end;
  end
  else if Dev[DeviceIndex].ImmADCs <> 0 then begin
    SetLength(ADBuf, Dev[DeviceIndex].ImmADCs);
    L := Dev[DeviceIndex].ImmADCs * 2;
    Result := GenericVendorRead(DeviceIndex, AUR_ADC_IMMEDIATE, 0, 0, L, @ADBuf[0]);
    StartChannel := 0;
    EndChannel := Dev[DeviceIndex].ImmADCs - 1;
    for I := 0 to 15 do Config[I] := $02; //0-5V
    Config[$13] := 0;
  end
  else begin
    Result := ERROR_BAD_TOKEN_TYPE;
  end;
end;

function ADC_GetScan(DeviceIndex: LongWord; pBuf: PWord): LongWord; cdecl;
var
  Config: array[0..21] of Byte;
  ADBuf: TWordArray;
  I, Ch, J, Tot: Integer;
  StartChannel, EndChannel: Byte;

begin
  Result := ADC_GetScan_Inner(DeviceIndex, Config, ADBuf, StartChannel, EndChannel, 0);
  if Result <> ERROR_SUCCESS then Exit;

  if Config[$13] <> 0 then begin
    I := 0;
    Inc(pBuf, StartChannel);
    for Ch := StartChannel to EndChannel do begin
      Tot := 0;
      for J := 1 to Config[$13] do
        Tot := Tot + ADBuf[I * (1 + Config[$13]) + J]
      ;
      pBuf^ := Round(Tot / Config[$13]);
      Inc(pBuf);
      Inc(I);
    end;
  end
  else begin
    I := 0;
    Inc(pBuf, StartChannel);
    for Ch := StartChannel to EndChannel do begin
      pBuf^ := ADBuf[I];
      Inc(pBuf);
      Inc(I);
    end;
  end;
end;

function ADC_GetScanV(DeviceIndex: LongWord; pBuf: PDouble): LongWord; cdecl;
var
  Config: array[0..21] of Byte;
  ADBuf: TWordArray;
  I, Ch, J, Tot: Integer;
  V: Double;
  StartChannel, EndChannel, RangeCode, RangeShift: Byte;

begin
  Result := ADC_GetScan_Inner(DeviceIndex, Config, ADBuf, StartChannel, EndChannel, 0);
  if Result <> ERROR_SUCCESS then Exit;

  RangeShift := Dev[DeviceIndex].RangeShift;
  if Config[$13] <> 0 then begin
    I := 0;
    Inc(pBuf, StartChannel);
    for Ch := StartChannel to EndChannel do begin
      RangeCode := Config[Ch shr RangeShift];

      Tot := 0;
      for J := 1 to Config[$13] do
        Tot := Tot + ADBuf[I * (1 + Config[$13]) + J]
      ;

      V := Tot / Config[$13] * (1 / 65536);
      if (RangeCode and 1) <> 0 then V := V * 2 - 1;
      if (RangeCode and 2) = 0 then V := V * 2;
      if (RangeCode and 4) = 0 then V := V * 5;
      pBuf^ := V;

      Inc(pBuf);
      Inc(I);
    end;
  end
  else begin
    I := 0;
    Inc(pBuf, StartChannel);
    for Ch := StartChannel to EndChannel do begin
      V := ADBuf[I] * (5 / 65536);
      pBuf^ := V;
      Inc(pBuf);
      Inc(I);
    end;
  end;
end;

function ADC_GetTrigScan(DeviceIndex: LongWord; pBuf: PWord; TimeoutMS: LongInt): LongWord; cdecl;
var
  Config: array[0..21] of Byte;
  ADBuf: TWordArray;
  I, Ch, J, Tot: Integer;
  StartChannel, EndChannel: Byte;

begin
  if TimeoutMS <= 0 then TimeoutMS := 1;
  Result := ADC_GetScan_Inner(DeviceIndex, Config, ADBuf, StartChannel, EndChannel, TimeoutMS);
  if Result <> ERROR_SUCCESS then Exit;

  if Config[$13] <> 0 then begin
    I := 0;
    Inc(pBuf, StartChannel);
    for Ch := StartChannel to EndChannel do begin
      Tot := 0;
      for J := 1 to Config[$13] do
        Tot := Tot + ADBuf[I * (1 + Config[$13]) + J]
      ;
      pBuf^ := Round(Tot / Config[$13]);
      Inc(pBuf);
      Inc(I);
    end;
  end
  else begin
    I := 0;
    Inc(pBuf, StartChannel);
    for Ch := StartChannel to EndChannel do begin
      pBuf^ := ADBuf[I];
      Inc(pBuf);
      Inc(I);
    end;
  end;
end;

function ADC_GetTrigScanV(DeviceIndex: LongWord; pBuf: PDouble; TimeoutMS: LongInt): LongWord; cdecl;
var
  Config: array[0..21] of Byte;
  ADBuf: TWordArray;
  I, Ch, J, Tot: Integer;
  V: Double;
  StartChannel, EndChannel, RangeCode, RangeShift: Byte;

begin
  if TimeoutMS <= 0 then TimeoutMS := 1;
  Result := ADC_GetScan_Inner(DeviceIndex, Config, ADBuf, StartChannel, EndChannel, TimeoutMS);
  if (Result <> ERROR_SUCCESS) and (Result <> ERROR_TIMEOUT) then Exit;

  RangeShift := Dev[DeviceIndex].RangeShift;
  if Config[$13] <> 0 then begin
    I := 0;
    Inc(pBuf, StartChannel);
    for Ch := StartChannel to EndChannel do begin
      RangeCode := Config[Ch shr RangeShift];

      Tot := 0;
      for J := 1 to Config[$13] do
        Tot := Tot + ADBuf[I * (1 + Config[$13]) + J]
      ;

      V := Tot / Config[$13] * (1 / 65536);
      if (RangeCode and 1) <> 0 then V := V * 2 - 1;
      if (RangeCode and 2) = 0 then V := V * 2;
      if (RangeCode and 4) = 0 then V := V * 5;
      pBuf^ := V;

      Inc(pBuf);
      Inc(I);
    end;
  end
  else begin
    I := 0;
    Inc(pBuf, StartChannel);
    for Ch := StartChannel to EndChannel do begin
      V := ADBuf[I] * (5 / 65536);
      pBuf^ := V;
      Inc(pBuf);
      Inc(I);
    end;
  end;
end;

function ADC_GetChannelV(DeviceIndex, ChannelIndex: LongWord; pBuf: PDouble): LongWord; cdecl;
var
  ScanData: array of Double;
  I: Integer;
  Chans: LongWord;

begin
  Result := Validate(DeviceIndex);
  if Result <> ERROR_SUCCESS then Exit;

  with Dev[DeviceIndex] do begin
    if bADCBulk then
      Chans := ADCMUXChannels
    else if ImmADCs <> 0 then
      Chans := ImmADCs
    else begin
      Result := ERROR_BAD_TOKEN_TYPE;
      Exit;
    end;

    if ChannelIndex >= Chans then begin
      Result := ERROR_INVALID_INDEX;
      Exit;
    end;

    SetLength(ScanData, Chans);
    for I := 0 to Chans - 1 do ScanData[I] := NAN;

    Result := ADC_GetScanV(DeviceIndex, @ScanData[0]);
    if Result <> ERROR_SUCCESS then Exit;

    pBuf^ := ScanData[ChannelIndex];
  end;
end;

function ADC_GetCalRefV(DeviceIndex, CalRefIndex: LongWord; var pRef: Double): LongWord; cdecl;
var
  oConfig, nConfig: array of Byte;
  L: LongWord;
  bBip: Boolean;
  ADData: array of Double;

begin
  try
    Result := ValidateAndEnsureOpen(DeviceIndex);
    if Result <> ERROR_SUCCESS then Exit;

    with Dev[DeviceIndex] do begin
      if not bADCBulk then begin
        Result := ERROR_BAD_TOKEN_TYPE;
        Exit;
      end;

      if ConfigBytes < 20 then begin
        Result := ERROR_BAD_TOKEN_TYPE;
        Exit;
      end;

      L := ConfigBytes;
      SetLength(oConfig, L);
      SetLength(nConfig, L);
      FillChar(oConfig[0], L, $00);
      FillChar(nConfig[0], L, $00);
      Result := ADC_GetConfig(DeviceIndex, @oConfig[0], L);
      if Result <> ERROR_SUCCESS then Exit;

      try
        bBip := (oConfig[$00] and 1) <> 0; //Select bip or unip ref based on how they've set channel 0.
        nConfig[$00] := oConfig[$00]; //Use their selected range.
        case CalRefIndex of
          0: if bBip then //Lo ref.
            nConfig[$10] := $05
          else
            nConfig[$10] := $01
          ;
          1: if bBip then //Hi ref.
            nConfig[$10] := $07
          else
            nConfig[$10] := $03
          ;
          else begin
            Result := ERROR_INVALID_INDEX;
            Exit;
          end;
        end;
        nConfig[$13] := $FF; //Max oversample, make sure we get a good read.
        nConfig[$12] := $00; //Scan one channel.
        if ConfigBytes > $14 then nConfig[$14] := $00; //Scan one channel.
        ADC_SetConfig(DeviceIndex, @nConfig[0], L);

        SetLength(ADData, ADCMUXChannels);
        Result := ADC_GetScanV(DeviceIndex, @ADData[0]);
        if Result <> ERROR_SUCCESS then Exit;
      finally
        L := ConfigBytes;
        ADC_SetConfig(DeviceIndex, @oConfig[0], L);
        SetLength(oConfig, 0);
        SetLength(nConfig, 0);
      end;

      pRef := ADData[0];
      //Result := ERROR_SUCCESS;
    end;
  except
    Result := ERROR_INTERNAL_ERROR;
  end;
end;



end.
 
